import logging
from pathlib import Path

import click

log = logging.getLogger(__name__)


@click.group(name='file')
def main():
    pass


@main.command()
@click.argument('path', type=click.Path(), nargs=-1)
def index(path):
    """
    Index a file.

    Consume STDIN if no arguments given. This is useful for pipelines like this:

        find . -type f -name '*.py' | puddl file index
    """
    # The path argument could use a better name, but the help text needs to be sane.
    # Using metavar makes things unnecessarily complicated. We want to keep backus naur.
    def _iter_paths():
        if not path:
            for line in click.get_text_stream('stdin'):
                yield Path(line.rstrip('\n'))
        else:
            for x in path:
                yield x

    from puddl.models import Device
    from .models import File
    for p in _iter_paths():
        log.info(f'indexing {p}')
        File.objects.upsert(
            device=Device.objects.me(),
            path=p
        )


@main.command()
def ls():
    from .models import File
    for f in File.objects.all():
        print(f.path)


@main.command('query')
@click.argument('query_string', type=click.STRING)
def q(query_string):
    """
    query by "path contains"
    """
    from puddl.contrib.file.models import File
    for x in File.objects.filter(path__contains=query_string):
        print(x.path)



MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'puddl.project.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'puddl.project.wsgi.application'


def logging_config(log_level='INFO', log_sql=False):
    result = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            # https://docs.python.org/3/library/logging.html#logrecord-attributes
            'simple': {
                'format': '%(levelname)s %(message)s',
            },
            'standard': {
                'format': '[%(asctime)s] %(levelname)s %(name)s %(message)s',
                'datefmt': "%Y-%m-%d %H:%M:%S"
            },
        },
        'handlers': {
            'console': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
        },
        'loggers': {
            # https://stackoverflow.com/a/31251707
            'puddl': {
                'handlers': ['console'],
                'level': log_level,
                'propagate': False,
            },
            'django': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': True,
            },
        }
    }
    if log_sql:
        result['loggers']['django.db.backends'] = {
            'level': 'DEBUG',
        }
    return result


LOGGING = logging_config()


AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME':
        'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

STATIC_URL = '/static/'


# Queue Stuff
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.db.DatabaseCache',
        'LOCATION': '_django_cache',
    }
}
# https://django-q.readthedocs.io/en/latest/configure.html
Q_CLUSTER = {
    'name': 'DjangORM',
    'orm': 'default',  # DB name
    'workers': 2,
    'timeout': 3600,  # scrapes take a while
    'retry': 3666,
}



import logging
from pathlib import Path

from django.db import models
from django.contrib.postgres.fields import JSONField
import filetype

from puddl.fields import PathField
from puddl.models import Datum, Device

log = logging.getLogger(__name__)


class FileManager(models.Manager):

    def upsert(self, device: Device, path: Path):
        path = Path(path)
        try:
            instance = self.get(device=device, path=path)
            log.debug(f'found {instance}')
        except self.model.DoesNotExist:
            instance = self.model()
            instance.device = device
            instance.path = path
            log.debug(f'created {instance}')
        instance.filetype = instance.guess_filetype()
        instance.stat = instance.get_stat()
        instance.save()
        return instance


class File(Datum):
    objects = FileManager()

    device = models.ForeignKey(Device, on_delete=models.CASCADE)
    path = PathField()

    filetype = JSONField(null=True)
    stat = JSONField()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['device', 'path'],
                name='unique_device_path',
            )
        ]

    @property
    def ssh_path(self):
        return f'{self.device.name}:{self.path}'

    def __str__(self):
        return self.ssh_path

    def guess_filetype(self):
        x = filetype.guess(str(self.path))
        if x is not None:
            return {
                "extension": x.extension,
                "mime": x.mime
            }

    def get_stat(self):
        s = self.path.stat()
        attrs = [a for a in dir(s) if a.startswith('st_')]
        return {a: getattr(s, a) for a in attrs}



# TODO rename me to config :>

import json
import logging
from pathlib import Path

log = logging.getLogger(__name__)


PUDDLRC = Path.home() / '.puddlrc'


def django_setup(debug=False, log_level='INFO', log_sql=False, skipconfigured=False,
                 **kwargs):
    """
    :param skipconfigured: explicitly silence Django "already configured" exception
    """
    from django.conf import settings
    from puddl.project.settings import logging_config
    if skipconfigured and settings.configured:
        return
    config = read_rcfile()
    if debug:
        log_level = 'DEBUG'
    # TODO get stuff from ~/.puddlrc
    result = dict(
        DEBUG=debug,
        LOGGING=logging_config(log_level, log_sql),
        # ALLOWED_HOSTS=...,
        SECRET_KEY='^6x*s4#x8!+zcv9j5h69h+=#9l#h)^u&q6dxx@5ir_+fr7p1)j',
        # if you add stuff here, maybe exclude it in puddl.cli.PluginsCommand
        INSTALLED_APPS=[
            'django.contrib.postgres',
            'puddl',
            'puddl.contrib.file',
        ] + config.get('plugins', []),
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.postgresql',
                'HOST': config['PGHOST'],
                'PORT': config['PGPORT'],
                'NAME': config['PGDATABASE'],
                'USER': config['PGUSER'],
                'PASSWORD': config['PGPASSWORD'],
            }
        },
    )
    result.update(kwargs)
    settings.configure(**result)
    import django
    django.setup()
    if settings.DEBUG:
        log.warning(f'debug mode enabled')
    return settings


def read_rcfile():
    try:
        with PUDDLRC.open() as f:
            return json.load(f)
    except FileNotFoundError:
        with PUDDLRC.open('w') as f:
            json.dump({}, f)
            return {}


def write_rcfile(config):
    with PUDDLRC.open('w') as f:
        json.dump(config, f, indent=2, sort_keys=True)



from pathlib import Path
from django.db import models


class PathField(models.TextField):
    """
    Pathlib FTW! Saves absolute paths to DB.

    >>> class MyModel(models.Model):
            path = PathField()
    >>> x = MyModel()
    >>> type(x.path)
    pathlib.Path
    """

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return Path(value)

    def to_python(self, value):
        if isinstance(value, str) or value is None:
            return value
        abspath = value.resolve()
        return str(abspath)



import logging

import click

log = logging.getLogger(__name__)
LOG_LEVELS = ['CRITICAL', 'FATAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG']


class PluginsCommand(click.Group):
    """
    Hook plugins.

    https://click.palletsprojects.com/en/7.x/commands/#custom-multi-commands
    """

    def list_commands(self, ctx):
        core_commands = set(super().list_commands(ctx))
        from puddl.conf import django_setup
        django_setup(skipconfigured=True)
        from django.apps import apps
        candidates = [app.label for app in apps.get_app_configs()]
        commands = set(x for x in candidates if x not in ['puddl', 'postgres'])
        # the label of an installed app must not overlap with puddl's core commands
        invalid_commands = sorted(core_commands & commands)
        if invalid_commands:
            log.error(f'invalid commands: {invalid_commands}')
            raise SystemExit(1)
        return sorted(core_commands | commands)

    def get_command(self, ctx, name):
        core_commands = super().list_commands(ctx)
        if name in core_commands:
            return globals()[name]
        from puddl.conf import django_setup
        django_setup(skipconfigured=True)
        from django.apps import apps
        x = apps.get_app_config(name)
        import importlib
        mod = importlib.util.resolve_name('.cli', x.module.__package__)
        entrypoint = importlib.import_module(mod).main
        if entrypoint.name != name:
            log.warning(f'plugin {name} should set "name={name}" in cli.main()')
            entrypoint.name = name
        return entrypoint


@click.group(cls=PluginsCommand)
@click.option('-d', '--debug/--no-debug')
@click.option('-l', '--log-level',
              type=click.Choice(LOG_LEVELS, case_sensitive=False),
              default='INFO')
@click.option('--log-sql/--no-log-sql')
@click.version_option()
def root(debug, log_level, log_sql):
    from puddl.conf import django_setup
    django_setup(debug, log_level, log_sql, skipconfigured=True)


@root.group()
def config():
    pass


@config.command()
@click.option('--from-key-value', is_flag=True)
def update(from_key_value):
    from puddl.conf import read_rcfile, write_rcfile
    config = read_rcfile()
    if from_key_value:
        updates = {}
        for line in click.get_text_stream('stdin').read().strip().split('\n'):
            k, v = line.split('=')
            updates[k] = v
        config.update(updates)
    write_rcfile(config)


@root.group()
def app():
    pass


@app.command('add')
@click.argument('package_path')
def add_app(package_path):
    from puddl.conf import read_rcfile, write_rcfile
    config = read_rcfile()
    if 'plugins' not in config:
        config['plugins'] = []
    tmp = config.get('plugins', []) + [package_path]
    config['plugins'] = list(sorted(set(tmp)))
    write_rcfile(config)


@app.command('ls')
def app_ls():
    from django.apps import apps
    xs = [f'{x.label} ({x.name})' for x in apps.get_app_configs()]
    print('\n'.join((sorted(xs))))


@root.group()
def db():
    pass


@db.command()
def health():
    from django.db import connection

    with connection.cursor() as cursor:
        cursor.execute('SELECT 1')
    log.info('database available')


@db.command()
def migrate():
    from django.core import management
    management.call_command('migrate')


@db.command()
def shell():
    from django.core import management
    management.call_command('dbshell')


def main():
    # this function is referenced in `setup.py`
    root(auto_envvar_prefix='PUDDL')


if __name__ == '__main__':
    # https://click.palletsprojects.com/en/7.x/options/#values-from-environment-variables
    main()
